# Copyright 2021-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time


class StateMachine:
    START = 'start'
    COMPLETE = 'complete'

    def __init__(self, initial_state):
        self.state = initial_state.get("state", self.START)
        self.complete = False
        self.start_time = initial_state.get("start_time", time.time())

    def advance(self):
        pass

    def toDict(self):
        return dict(
            state=self.state,
            start_time=self.start_time,
        )

    def fail(self):
        self.state = self.COMPLETE
        self.complete = True


class Instance:
    """Represents a cloud instance

    This class is used by the State Machine Driver classes to
    represent a standardized version of a remote cloud instance.
    Implement this class in your driver, override the :py:meth:`load`
    method, and supply as many of the fields as possible.

    The following attributes are required:

    * ready: bool (whether the instance is ready)
    * deleted: bool (whether the instance is in a deleted state)
    * interface_ip: str
    * metadata: dict

    The following are optional:

    * public_ipv4: str
    * public_ipv6: str
    * private_ipv4: str
    * cloud: str
    * az: str
    * region: str
    * host_id: str
    * driver_data: any
    * slot: int

    And the following are even more optional (as they are usually
    already set from the image configuration):

    * username: str
    * python_path: str
    * shell_type: str
    * connection_port: str
    * connection_type: str
    * host_keys: [str]

    This is extremely optional, in fact, it's difficult to imagine
    that it's useful for anything other than the metastatic driver:

    * node_attributes: dict

    """
    def __init__(self):
        self.ready = False
        self.deleted = False
        self.public_ipv4 = None
        self.public_ipv6 = None
        self.private_ipv4 = None
        self.private_ipv6 = None
        self.interface_ip = None
        self.cloud = None
        self.az = None
        self.region = None
        self.host_id = None
        self.metadata = {}
        self.driver_data = None

    @property
    def external_id(self):
        """Return a string representation of external ID(s) for this instance,
        for use in logging"""
        raise NotImplementedError()

    def __repr__(self):
        state = []
        if self.ready:
            state.append('ready')
        if self.deleted:
            state.append('deleted')
        state = ' '.join(state)
        return '<{klass} {external_id} {state}>'.format(
            klass=self.__class__.__name__,
            external_id=self.external_id,
            state=state)

    def getQuotaInformation(self):
        """Return quota information about this instance.

        :returns: A :py:class:`QuotaInformation` object.
        """
        raise NotImplementedError()


class Resource:
    """Represents a cloud resource

    This could be an instance, a disk, a floating IP, or anything
    else.  It is used by the driver to detect leaked resources so the
    adapter can clean them up.

    The `type` attribute should be an alphanumeric string suitable for
    inclusion in a statsd metric name.

    The `metadata` attribute is a dictionary of key/value pairs
    initially supplied by the driver to the adapter when an instance
    or image was created.  This is used by the driver to detect leaked
    resources.  The adapter may add any other information to this
    instance for its own bookeeping (resource type, id, etc).

    The 'plural_metric_name' attribute is set in the constructor
    automatically; override this value if necessary.

    :param str type: The type of resource.
    :param dict metadata: A dictionary of metadata for the resource.

    """

    def __init__(self, metadata, type):
        self.type = type
        self.plural_metric_name = type + 's'
        self.metadata = metadata

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.type} {self.metadata}>'

    @property
    def unique_id(self):
        """Return a unique id string for this resource"""
        raise NotImplementedError()



# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from zuul.lib.logutil import get_annotated_logger

import cherrypy
from cherrypy._cplogging import LogManager


class ZuulCherrypyLogManager(LogManager):
    @property
    def access_log(self):
        request = cherrypy.serving.request
        if not hasattr(request, 'zuul_request_id'):
            request.zuul_request_id = uuid.uuid4().hex
        return get_annotated_logger(self._access_log,
                                    request=request.zuul_request_id)

    @access_log.setter
    def access_log(self, attr):
        self._access_log = attr

    @property
    def error_log(self):
        request = cherrypy.serving.request
        if not hasattr(request, 'zuul_request_id'):
            request.zuul_request_id = uuid.uuid4().hex
        return get_annotated_logger(self._error_log,
                                    request=request.zuul_request_id)

    @error_log.setter
    def error_log(self, attr):
        self._error_log = attr



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Run this to clean up leaked lock entries in the blob store (but only
if there are too many for Zuul to deal with on its own).

Run this command to get a list of potential keys:

  ./bin/zkSnapShotToolkit.sh /data/version-2/snapshot.XXX \
    | grep /zuul/cache/blob/lock

Pass the result as input to this script.
"""

import argparse
import sys

import kazoo.client
from kazoo.exceptions import NotEmptyError


class Cleanup:
    def __init__(self, args):
        kwargs = {}
        if args.cert:
            kwargs['use_ssl'] = True
            kwargs['keyfile'] = args.key
            kwargs['certfile'] = args.cert
            kwargs['ca'] = args.ca
        self.client = kazoo.client.KazooClient(args.host, **kwargs)
        self.client.start()

    def run(self):
        prefix = '/zuul/cache/blob/lock/'

        for line in sys.stdin:
            line = line.strip()
            if not line.startswith(prefix):
                continue
            key = line[len(prefix):]

            blob_path = f"/zuul/cache/blob/data/{key[0:2]}/{key}"
            lock_path = f"/zuul/cache/blob/lock/{key}"
            if self.client.exists(blob_path):
                continue
            try:
                self.client.delete(lock_path)
            except NotEmptyError:
                pass


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('host', help='ZK host string')
    parser.add_argument('--cert', help='Path to TLS certificate')
    parser.add_argument('--key', help='Path to TLS key')
    parser.add_argument('--ca', help='Path to TLS CA cert')
    args = parser.parse_args()

    clean = Cleanup(args)
    clean.run()



import gzip
import os
import re
import yaml


def get_log_age(path):
    filename = os.path.basename(path)
    parts = filename.split('.')
    if len(parts) < 4:
        return 0
    else:
        return int(parts[2])


class LogScraper(object):
    # Example log line
    # 2018-10-26 16:14:47,527 INFO zuul.nodepool: Nodeset <NodeSet two-centos-7-nodes [<Node 0000058431 ('primary',):centos-7>, <Node 0000058468 ('secondary',):centos-7>]> with 2 nodes was in use for 6241.08082151413 seconds for build <Build 530c4ca7af9e44dcb535e7074258e803 of tripleo-ci-centos-7-scenario008-multinode-oooq-container voting:False on <Worker ze05.openstack.org>> for project openstack/tripleo-quickstart-extras  # noqa
    r = re.compile(r'(?P<timestamp>\d+-\d+-\d+ \d\d:\d\d:\d\d,\d\d\d) INFO zuul.nodepool: Nodeset <.*> with (?P<nodes>\d+) nodes was in use for (?P<secs>\d+(.[\d\-e]+)?) seconds for build <Build \w+ of (?P<job>[^\s]+) voting:\w+ on .* for project (?P<repos>[^\s]+)')  # noqa

    def __init__(self):
        self.repos = {}
        self.sorted_repos = []
        self.jobs = {}
        self.sorted_jobs = []
        self.total_usage = 0.0
        self.projects = {}
        self.sorted_projects = []
        self.start_time = None
        self.end_time = None

    def scrape_file(self, fn):
        if fn.endswith('.gz'):
            open_f = gzip.open
        else:
            open_f = open
        with open_f(fn, 'rt') as f:
            for line in f:
                if 'nodes was in use for' in line:
                    m = self.r.match(line)
                    if not m:
                        continue
                    g = m.groupdict()
                    repo = g['repos']
                    secs = float(g['secs'])
                    nodes = int(g['nodes'])
                    job = g['job']
                    if not self.start_time:
                        self.start_time = g['timestamp']
                    self.end_time = g['timestamp']
                    if repo not in self.repos:
                        self.repos[repo] = {}
                        self.repos[repo]['total'] = 0.0
                    node_time = nodes * secs
                    self.total_usage += node_time
                    self.repos[repo]['total'] += node_time
                    if job not in self.jobs:
                        self.jobs[job] = 0.0
                    if job not in self.repos[repo]:
                        self.repos[repo][job] = 0.0
                    self.jobs[job] += node_time
                    self.repos[repo][job] += node_time

    def list_log_files(self, path='/var/log/zuul'):
        ret = []
        entries = os.listdir(path)
        prefix = os.path.join(path, 'zuul.log')
        for entry in entries:
            entry = os.path.join(path, entry)
            if os.path.isfile(entry) and entry.startswith(prefix):
                ret.append(entry)
        ret.sort(key=get_log_age, reverse=True)
        return ret

    def sort_repos(self):
        for repo in self.repos:
            self.sorted_repos.append((repo, self.repos[repo]['total']))

        self.sorted_repos.sort(key=lambda x: x[1], reverse=True)

    def sort_jobs(self):
        for job, usage in self.jobs.items():
            self.sorted_jobs.append((job, usage))

        self.sorted_jobs.sort(key=lambda x: x[1], reverse=True)

    def calculate_project_usage(self):
        '''Group usage by logical project/effort

        It is often the case that a single repo doesn't capture the work
        of a logical project or effort. If this is the case in your situation
        you can create a projects.yaml file that groups together repos
        under logical project names to report usage by that logical grouping.

        The projects.yaml should be in your current directory and have this
        format:

          project_name:
            deliverables:
              logical_deliverable_name:
                repos:
                  - repo1
                  - repo2

          project_name2:
            deliverables:
              logical_deliverable_name2:
                repos:
                  - repo3
                  - repo4
        '''
        if not os.path.exists('projects.yaml'):
            return self.sorted_projects
        with open('projects.yaml') as f:
            y = yaml.load(f)
        for name, v in y.items():
            self.projects[name] = 0.0
            for deliverable in v['deliverables'].values():
                for repo in deliverable['repos']:
                    if repo in self.repos:
                        self.projects[name] += self.repos[repo]['total']

        for project, usage in self.projects.items():
            self.sorted_projects.append((project, usage))

        self.sorted_projects.sort(key=lambda x: x[1], reverse=True)


scraper = LogScraper()
for fn in scraper.list_log_files():
    scraper.scrape_file(fn)

print('For period from %s to %s' % (scraper.start_time, scraper.end_time))
print('Total node time used: %.2fs' % scraper.total_usage)
print()

scraper.calculate_project_usage()
if scraper.sorted_projects:
    print('Top 20 logical projects by resource usage:')
    for project, total in scraper.sorted_projects[:20]:
        percentage = (total / scraper.total_usage) * 100
        print('%s: %.2fs, %.2f%%' % (project, total, percentage))
    print()

scraper.sort_repos()
print('Top 20 repos by resource usage:')
for repo, total in scraper.sorted_repos[:20]:
    percentage = (total / scraper.total_usage) * 100
    print('%s: %.2fs, %.2f%%' % (repo, total, percentage))
print()

scraper.sort_jobs()
print('Top 20 jobs by resource usage:')
for job, total in scraper.sorted_jobs[:20]:
    percentage = (total / scraper.total_usage) * 100
    print('%s: %.2fs, %.2f%%' % (job, total, percentage))
print()



# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os
import time

import boto3
from moto import mock_aws

import tests.base
from tests.base import (
    ZuulTestCase,
    iterate_timeout,
    okay_tracebacks,
    simple_layout,
)


FIXTURE_DIR = os.path.join(tests.base.FIXTURE_DIR, 'gerrit')


def serialize(event):
    return json.dumps(event).encode('utf8')


class TestGerritEventSourceAWSKinesis(ZuulTestCase):
    config_file = 'zuul-gerrit-awskinesis.conf'
    mock_aws = mock_aws()

    def setUp(self):
        self.mock_aws.start()

        self.kinesis_client = boto3.client('kinesis', region_name='us-west-2')
        self.kinesis_client.create_stream(
            StreamName='gerrit',
            ShardCount=4,
            StreamModeDetails={
                'StreamMode': 'ON_DEMAND'
            }
        )
        self.addCleanup(self.mock_aws.stop)
        super().setUp()

    @simple_layout('layouts/simple.yaml')
    def test_kinesis(self):
        listener = self.fake_gerrit.event_thread

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')

        self.kinesis_client.put_record(
            StreamName='gerrit',
            Data=serialize(A.getPatchsetCreatedEvent(1)),
            PartitionKey='whatever',
        )

        for _ in iterate_timeout(60, 'wait for event'):
            if listener._event_count == 1:
                break
            time.sleep(0.2)
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1')
        ])
        self.assertEqual(A.reported, 1, "A should be reported")

        # Stop the listener
        listener.stop()
        listener.join()

        # Add new gerrit events while we are "offline"
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        self.kinesis_client.put_record(
            StreamName='gerrit',
            Data=serialize(B.getPatchsetCreatedEvent(1)),
            PartitionKey='whatever',
        )

        # Restart the listener
        listener.init()
        listener.start()

        for _ in iterate_timeout(60, 'wait for caught up'):
            if all(listener._caught_up.values()):
                break
            time.sleep(0.2)
        self.waitUntilSettled()

        # Make sure we don't reprocess old events (change A), but do
        # see new events (change B)
        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
            dict(name='check-job', result='SUCCESS', changes='2,1'),
        ])
        self.assertEqual(A.reported, 1, "A should be reported")
        self.assertEqual(B.reported, 1, "B should be reported")

    @simple_layout('layouts/simple.yaml')
    @okay_tracebacks("invalid literal for int() with base 10: 'nope'")
    def test_kinesis_bad_checkpoint(self):
        listener = self.fake_gerrit.event_thread

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')

        self.kinesis_client.put_record(
            StreamName='gerrit',
            Data=serialize(A.getPatchsetCreatedEvent(1)),
            PartitionKey='whatever',
        )

        for _ in iterate_timeout(60, 'wait for event'):
            if listener._event_count == 1:
                break
            time.sleep(0.2)
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1')
        ])
        self.assertEqual(A.reported, 1, "A should be reported")

        # Stop the listener
        listener.stop()
        listener.join()

        # Corrupt the checkpoint
        for cp in listener.checkpoints.values():
            cp.set("nope")

        # Add new gerrit events while we are "offline"
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        self.kinesis_client.put_record(
            StreamName='gerrit',
            Data=serialize(B.getPatchsetCreatedEvent(1)),
            PartitionKey='whatever',
        )

        # Restart the listener
        listener.init()
        listener.start()

        for _ in iterate_timeout(60, 'wait for caught up'):
            if all(listener._caught_up.values()):
                break
            time.sleep(0.2)
        self.waitUntilSettled()

        # Make sure we don't reprocess old events (change A),
        # and also that we missed change B because of the corruption
        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
        ])
        self.assertEqual(A.reported, 1, "A should be reported")
        self.assertEqual(B.reported, 0, "B should not be reported")

        # Poke B again to make sure we get new events
        self.kinesis_client.put_record(
            StreamName='gerrit',
            Data=serialize(B.getPatchsetCreatedEvent(1)),
            PartitionKey='whatever',
        )

        for _ in iterate_timeout(60, 'wait for event'):
            if listener._event_count == 2:
                break
            time.sleep(0.2)
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
            dict(name='check-job', result='SUCCESS', changes='2,1'),
        ])
        self.assertEqual(A.reported, 1, "A should be reported")
        self.assertEqual(B.reported, 1, "B should be reported")



#!/usr/bin/env python
# Copyright 2014-2015 Antoine "hashar" Musso
# Copyright 2014-2015 Wikimedia Foundation Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# pylint: disable=locally-disabled, invalid-name

"""
Zuul references cleaner.

Clear up references under /refs/zuul/ by inspecting the age of the commit the
reference points to.  If the commit date is older than a number of days
specificed by --until, the reference is deleted from the git repository.

Use --dry-run --verbose to finely inspect the script behavior.
"""

import argparse
import git
import logging
import time
import sys

NOW = int(time.time())
DEFAULT_DAYS = 360
ZUUL_REF_PREFIX = 'refs/zuul/'

parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
parser.add_argument('--until', dest='days_ago', default=DEFAULT_DAYS, type=int,
                    help='references older than this number of day will '
                         'be deleted. Default: %s' % DEFAULT_DAYS)
parser.add_argument('-n', '--dry-run', dest='dryrun', action='store_true',
                    help='do not delete references')
parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                    help='set log level from info to debug')
parser.add_argument('gitrepo', help='path to a Zuul git repository')
args = parser.parse_args()

logging.basicConfig()
log = logging.getLogger('zuul-clear-refs')
if args.verbose:
    log.setLevel(logging.DEBUG)
else:
    log.setLevel(logging.INFO)

try:
    repo = git.Repo(args.gitrepo)
except git.exc.InvalidGitRepositoryError:
    log.error("Invalid git repo: %s" % args.gitrepo)
    sys.exit(1)

for ref in repo.references:

    if not ref.path.startswith(ZUUL_REF_PREFIX):
        continue
    if type(ref) is not git.refs.reference.Reference:
        # Paranoia: ignore heads/tags/remotes ..
        continue

    try:
        commit_ts = ref.commit.committed_date
    except LookupError:
        # GitPython does not properly handle PGP signed tags
        log.exception("Error in commit: %s, ref: %s. Type: %s",
                      ref.commit, ref.path, type(ref))
        continue

    commit_age = int((NOW - commit_ts) / 86400)  # days
    log.debug(
        "%s at %s is %3s days old",
        ref.commit,
        ref.path,
        commit_age,
    )
    if commit_age > args.days_ago:
        if args.dryrun:
            log.info("Would delete old ref: %s (%s)", ref.path, ref.commit)
        else:
            log.info("Deleting old ref: %s (%s)", ref.path, ref.commit)
            ref.delete(repo, ref.path)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc


class BaseExecutionContext(object, metaclass=abc.ABCMeta):
    """The execution interface returned by a wrapper.

    Wrapper drivers return instances which implement this interface.

    It is used to hold information and aid in the execution of a
    single command.

    """

    @abc.abstractmethod
    def getPopen(self, **kwargs):
        """Create and return a subprocess.Popen factory wrapped however the
        driver sees fit.

        This method is required by the interface

        :arg dict kwargs: key/values for use by driver as needed

        :returns: a callable that takes the same args as subprocess.Popen
        :rtype: Callable
        """
        pass

    @abc.abstractmethod
    def getNamespacePids(self, proc):
        """Given a Popen object, return the namespace and a list of host-side
        process ids in proc's first child namespace.

        :arg Popen proc: The Popen object that is the parent of the namespace.

        :returns: A tuple of the namespace id and all of the process
            ids in the namespace.  The namespace id is None if no namespace
            was found.
        :rtype: tuple(int, list[int])

        """
        pass



# Copyright 2017 Red Hat, Inc.
# Copyright 2021-2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import signal
import sys

import zuul.cmd
import zuul.model
import zuul.web
import zuul.driver.sql
import zuul.driver.github
import zuul.lib.auth


class WebServer(zuul.cmd.ZuulDaemonApp):
    app_name = 'web'
    app_description = 'A standalone Zuul web server.'

    def createParser(self):
        parser = super().createParser()
        self.addSubCommands(parser, zuul.web.COMMANDS)
        return parser

    def exit_handler(self, signum, frame):
        self.web.stop()

    def _run(self):
        info = zuul.model.WebInfo.fromConfig(self.config)

        # Validate config here before we spin up the ZuulWeb object
        for conn_name, connection in self.connections.connections.items():
            try:
                connection.validateWebConfig(self.config, self.connections)
            except Exception:
                self.log.exception("Error validating config")
                sys.exit(1)

        try:
            self.web = zuul.web.ZuulWeb(
                config=self.config,
                info=info,
                connections=self.connections,
                authenticators=self.authenticators,
            )
            self.connections.load(self.web.zk_client,
                                  self.web.component_registry)
        except Exception:
            self.log.exception("Error creating ZuulWeb:")
            sys.exit(1)

        signal.signal(signal.SIGUSR1, self.exit_handler)
        signal.signal(signal.SIGTERM, self.exit_handler)

        self.log.info('Zuul Web Server starting')
        self.web.start()

        try:
            self.web.join()
        except KeyboardInterrupt:
            print("Ctrl + C: asking process to exit nicely...\n")
            self.exit_handler(signal.SIGINT, None)
            self.web.join()

        self.log.info("Zuul Web Server stopped")

    def configure_authenticators(self):
        self.authenticators = zuul.lib.auth.AuthenticatorRegistry()
        self.authenticators.configure(self.config)

    def run(self):
        self.handleCommands()

        self.setup_logging('web', 'log_config')
        self.log = logging.getLogger("zuul.WebServer")

        try:
            self.configure_connections(database=True, sources=True,
                                       triggers=True, reporters=True,
                                       providers=True)
            self.configure_authenticators()
            self._run()
        except Exception:
            self.log.exception("Exception from WebServer:")


def main():
    WebServer().main()



# Copyright 2014 Rackspace Australia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import time

from zuul import model


class BaseSource(object, metaclass=abc.ABCMeta):
    """Base class for sources.

    A source class gives methods for fetching and updating changes. Each
    pipeline must have (only) one source. It is the canonical provider of the
    change to be tested.

    Defines the exact public methods that must be supplied."""

    def __init__(self, driver, connection, canonical_hostname, config=None):
        self.driver = driver
        self.connection = connection
        self.canonical_hostname = canonical_hostname
        self.config = config or {}

    @abc.abstractmethod
    def getRefSha(self, project, ref):
        """Return a sha for a given project ref."""

    @abc.abstractmethod
    def isMerged(self, change, head=None):
        """Determine if change is merged.

        If head is provided the change is checked if it is at head.

        Return False, True, or the merged commit sha if available.
        """

    @abc.abstractmethod
    def canMerge(self, change, allow_needs, event=None, allow_refresh=False):
        """Determine if change can merge.

        change: The change to check for mergeability
        allow_needs: The statuses/votes that are allowed to be missing on a
                     change, typically the votes the pipeline would set itself
                     before attempting to merge
        event: event information for log annotation
        allow_refresh: Allow refreshing of cached volatile data that cannot be
                       reliably kept up to date using events.
        """

    def postConfig(self):
        """Called after configuration has been processed."""

    @abc.abstractmethod
    def getChangeKey(self, event):
        """Get a ChangeKey from a ChangeManagementEvent or TriggerEvent"""

    @abc.abstractmethod
    def getChange(self, change_key, refresh=False, event=None):
        """Get the change represented by a change_key

        This method is called very frequently, and should generally
        return quickly.  The connection is expected to cache change
        objects and automatically update them as related events are
        received.

        The event is optional, and if present may be used to annotate
        log entries and supply additional information about the change
        if a refresh is necessary.

        If the change key does not correspond to this source, return
        None.

        """

    @abc.abstractmethod
    def getChangeByURL(self, url, event):
        """Get the change corresponding to the supplied URL.

        The URL may may not correspond to this source; if it doesn't,
        or there is no change at that URL, return None.

        """

    def getChangeByURLWithRetry(self, url, event):
        for x in range(3):
            # We retry this as we are unlikely to be able to report back
            # failures if our source is broken, but if we can get the
            # info on subsequent requests we can continue to do the
            # requested job work.
            try:
                return self.getChangeByURL(url, event)
            except Exception:
                # Note that if the change isn't found dep is None.
                # We do not raise in that case and do not need to handle it
                # here.
                retry = x != 2 and " Retrying" or ""
                self.log.exception("Failed to retrieve dependency %s.%s",
                                   url, retry)
                if retry:
                    time.sleep(1)
                else:
                    raise

    @abc.abstractmethod
    def getChangesDependingOn(self, change, projects, tenant):
        """Return changes which depend on changes at the supplied URIs.

        Search this source for changes which depend on the supplied
        change.  Generally the Change.uris attribute should be used to
        perform the search, as it contains a list of URLs without the
        scheme which represent a single change

        If the projects argument is None, search across all known
        projects.  If it is supplied, the search may optionally be
        restricted to only those projects.

        The tenant argument can be used by the source to limit the
        search scope.
        """

    def useDependenciesByTopic(self):
        """Return whether the source uses topic dependencies

        If a source treats changes in a topic as a dependency cycle,
        this will return True.

        This is only implemented by the Gerrit driver, however if
        other systems have a similar "topic" functionality, it could
        be added to other drivers.
        """
        return False

    def getChangesByTopic(self, topic, event=None):
        """Return changes in the same topic.

        This should return changes under the same topic, as well as
        changes under the same topic of any git-dependent changes,
        recursively.

        This is only implemented by the Gerrit driver, however if
        other systems have a similar "topic" functionality, it could
        be added to other drivers.
        """

        return []

    @abc.abstractmethod
    def getProjectOpenChanges(self, project):
        """Get the open changes for a project."""

    def getProjectDefaultMergeMode(self, project, valid_modes=None):
        """Return the default merge mode for this project.

        If users do not specify the merge mode for a project, this
        mode will be used.  It may be a driver-specific default,
        or the driver may use data from the remote system to provide
        a project-specific default.

        Since the driver's default mode might change to a mode not yet
        supported by the active project configuration, the driver
        should select the most appropriate default base on the list of
        valid modes.
        """
        return 'merge'

    @abc.abstractmethod
    def getGitUrl(self, project):
        """Get the git url for a project."""

    def getRetryTimeout(self, project):
        """Get the retry timeout for a project in seconds.

        This is used by the mergers to potentially increase the number
        of git fetch retries before giving up.  Return None to use the
        default.
        """
        return None

    @abc.abstractmethod
    def getProject(self, name):
        """Get a project."""

    @abc.abstractmethod
    def getProjectBranches(self, project, tenant, min_ltime=-1):
        """Get branches for a project

        This method is called very frequently, and should generally
        return quickly.  The connection is expected to cache branch
        lists for all projects queried, and further, to automatically
        clear or update that cache when it observes branch creation or
        deletion events.

        """

    def getProjectMergeModes(self, project, tenant, min_ltime=-1):
        """Get supported merge modes for a project

        This method is called very frequently, and should generally
        return quickly.  The connection is expected to cache merge
        modes for all projects queried.

        The default implementation indicates that all merge modes are
        supported.

        """

        return model.ALL_MERGE_MODES

    def getProjectDefaultBranch(self, project, tenant, min_ltime=-1):
        """Return the default branch for this project.

        If users do not specify the default branch for a project, this
        mode will be used.  It may be a driver-specific default, or
        the driver may use data from the remote system to provide a
        project-specific default.

        This method is called very frequently, and should generally
        return quickly.  The connection is expected to cache default
        branches for all projects queried.

        """

        return 'master'

    @abc.abstractmethod
    def getProjectBranchSha(self, project, branch_name):
        """Return the SHA of the specified branch"""

    @abc.abstractmethod
    def getProjectBranchCacheLtime(self):
        """Return the current ltime of the project branch cache."""

    @abc.abstractmethod
    def getRequireFilters(self, config, error_accumulator):
        """Return a list of ChangeFilters for the scheduler to match against.
        """

    @abc.abstractmethod
    def getRejectFilters(self, config, error_accumulator):
        """Return a list of ChangeFilters for the scheduler to match against.
        """

    def setChangeAttributes(self, change, **attrs):
        """"Set the provided attributes on the given change.

        This method must be used when modifying attributes of a change
        outside the driver context. The driver needs to make sure that
        the change is also reflected in the cache in Zookeeper.
        """
        # TODO (swestphahl): Remove this workaround after all drivers
        # have been converted to use a Zookeeper backed changed cache.
        for name, value in attrs.items():
            setattr(change, name, value)



#!/usr/bin/env python

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from zuul.lib import encryption
import zuul.configloader
import zuul.model


DESCRIPTION = """Decrypt a Zuul secret.
"""


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('private_key',
                        help="The path to the private key")
    parser.add_argument('file',
                        help="The YAML file with secrets")
    args = parser.parse_args()

    (private_secrets_key, public_secrets_key) = \
        encryption.deserialize_rsa_keypair(open(args.private_key, 'rb').read())
    parser = zuul.configloader.SecretParser(None)
    sc = zuul.model.SourceContext(None, 'project', None, 'master',
                                  'path', False)

    data = zuul.configloader.safe_load_yaml(open(args.file).read(), sc)
    for element in data:
        if 'secret' not in element:
            continue
        s = element['secret']
        secret = parser.fromYaml(s)
        print(secret.name)
        print(secret.decrypt(private_secrets_key).secret_data)


if __name__ == '__main__':
    main()



# Copyright 2011 OpenStack, LLC.
# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import paramiko
import pprint
import select
import threading
import time

from zuul.zk.event_queues import EventReceiverElection


class GerritSSHEventListener(threading.Thread):
    log = logging.getLogger("zuul.GerritConnection.ssh")
    poll_timeout = 500

    def __init__(self, gerrit_connection, connection_config):
        threading.Thread.__init__(self)
        self.username = connection_config.get('user')
        self.keyfile = connection_config.get('sshkey', None)
        server = connection_config.get('server')
        self.hostname = connection_config.get('ssh_server', server)
        self.port = int(connection_config.get('port', 29418))
        self.gerrit_connection = gerrit_connection
        self._stop_event = threading.Event()
        self.watcher_election = EventReceiverElection(
            gerrit_connection.sched.zk_client,
            gerrit_connection.connection_name,
            "watcher")
        self.keepalive = int(connection_config.get('keepalive', 60))
        self._stopped = False

    def _read(self, fd):
        while True:
            l = fd.readline()
            data = json.loads(l)
            self.log.debug("Received data from Gerrit event stream: \n%s" %
                           pprint.pformat(data))
            self.gerrit_connection.addEvent(data)
            # Continue until all the lines received are consumed
            if fd._pos == fd._realpos:
                break

    def _listen(self, stdout, stderr):
        poll = select.poll()
        poll.register(stdout.channel)
        while not self._stopped and self.watcher_election.is_still_valid():
            ret = poll.poll(self.poll_timeout)
            if not self.watcher_election.is_still_valid():
                return
            for (fd, event) in ret:
                if fd == stdout.channel.fileno():
                    if event == select.POLLIN:
                        self._read(stdout)
                    else:
                        raise Exception("event on ssh connection")

    def _run(self):
        try:
            client = paramiko.SSHClient()
            client.load_system_host_keys()
            client.set_missing_host_key_policy(paramiko.WarningPolicy())
            # SSH banner, handshake, and auth timeouts default to 15
            # seconds, so we only set the socket timeout here.
            client.connect(self.hostname,
                           username=self.username,
                           port=self.port,
                           key_filename=self.keyfile,
                           timeout=self.gerrit_connection.ssh_timeout)
            transport = client.get_transport()
            transport.set_keepalive(self.keepalive)

            # We don't set a timeout for this since it is never
            # expected to complete.  We rely on ssh and tcp keepalives
            # to detected a down connection.
            stdin, stdout, stderr = client.exec_command("gerrit stream-events")

            self._listen(stdout, stderr)

            if not stdout.channel.exit_status_ready():
                # The stream-event is still running but we are done polling
                # on stdout most likely due to being asked to stop.
                # Try to stop the stream-events command sending Ctrl-C
                stdin.write("\x03")
                time.sleep(.2)
                if not stdout.channel.exit_status_ready():
                    # we're still not ready to exit, lets force the channel
                    # closed now.
                    stdout.channel.close()
            ret = stdout.channel.recv_exit_status()
            self.log.debug("SSH exit status: %s" % ret)

            if ret and ret not in [-1, 130]:
                raise Exception("Gerrit error executing stream-events")
        finally:
            # If we don't close on exceptions to connect we can leak the
            # connection and DoS Gerrit.
            client.close()

    def run(self):
        while not self._stopped:
            try:
                self.watcher_election.run(self._run)
            except Exception:
                self.log.exception("Exception on ssh event stream with %s:",
                                   self.gerrit_connection.connection_name)
                self._stop_event.wait(5)

    def stop(self):
        self.log.debug("Stopping watcher")
        self._stopped = True
        self._stop_event.set()
        self.watcher_election.cancel()



# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import time
from uuid import uuid4

import zuul.executor.common
from zuul.lib.logutil import get_annotated_logger
from zuul.model import (
    Build,
    BuildCompletedEvent,
    BuildRequest,
    BuildStartedEvent,
    PRIORITY_MAP,
)
from zuul.zk.event_queues import PipelineResultEventQueue
from zuul.zk.executor import ExecutorApi
from zuul.zk.exceptions import JobRequestNotFound
import zuul.lib.tracing as tracing

from kazoo.exceptions import BadVersionError
from opentelemetry import trace


class ExecutorClient(object):
    log = logging.getLogger("zuul.ExecutorClient")
    _executor_api_class = ExecutorApi

    def __init__(self, config, sched):
        self.config = config
        self.sched = sched

        self.executor_api = self._executor_api_class(self.sched.zk_client)
        self.result_events = PipelineResultEventQueue.createRegistry(
            self.sched.zk_client
        )

    def stop(self):
        self.log.debug("Stopping")
        self.executor_api.stop()

    def getImageFormats(self, job, item):
        if not job.image_build_name:
            return None
        formats = set()
        for provider in item.manager.tenant.layout.providers.values():
            for image in provider.images.values():
                if image.name == job.image_build_name:
                    formats.add(image.format)
        return formats

    def execute(self, job, nodes, item, pipeline, executor_zone,
                dependent_changes=[], merger_items=[]):
        log = get_annotated_logger(self.log, item.event)
        tracer = trace.get_tracer("zuul")
        uuid = str(uuid4().hex)
        manager = item.manager
        log.info(
            "Execute job %s (uuid: %s) on nodes %s for %s "
            "with dependent changes %s",
            job, uuid, nodes, item, dependent_changes)

        params = zuul.executor.common.construct_build_params(
            uuid, self.sched.connections,
            job, item, pipeline, dependent_changes, merger_items,
            redact_secrets_and_keys=False)

        # Pass webroot to the executor for generating oidc token
        web_root = manager.tenant.web_root
        if web_root:
            params["zuul_root_url"] = web_root.split("/t/")[0].rstrip("/")
        # TODO: deprecate and remove this variable?
        params["zuul"]["_inheritance_path"] = list(job.inheritance_path)

        if job.image_build_name:
            params["zuul"]["image_build_name"] = job.image_build_name
        if image_formats := self.getImageFormats(job, item):
            params["zuul"]["image_formats"] = list(image_formats)

        semaphore_handler = manager.tenant.semaphore_handler
        params['semaphore_handle'] = semaphore_handler.getSemaphoreHandle(
            item, job)

        parent_span = tracing.restoreSpan(item.current_build_set.span_info)
        execute_time = time.time()
        with trace.use_span(parent_span):
            build_span = tracer.start_span("Build", start_time=execute_time)
        build_span_info = tracing.getSpanInfo(build_span)
        build = Build.new(
            manager.current_context,
            job=job,
            build_set=item.current_build_set,
            uuid=uuid,
            execute_time=execute_time,
            span_info=build_span_info,
            zuul_event_id=item.event.zuul_event_id,
        )

        log.debug("Adding build %s of job %s to item %s",
                  build, job, item)
        item.addBuild(build)

        if job.name == 'noop':
            data = {"start_time": time.time()}
            started_event = BuildStartedEvent(
                build.uuid, build.build_set.uuid, job.uuid,
                None, data, zuul_event_id=build.zuul_event_id)
            self.result_events[manager.tenant.name][manager.pipeline.name].put(
                started_event
            )

            result = {"result": "SUCCESS", "end_time": time.time()}
            completed_event = BuildCompletedEvent(
                build.uuid, build.build_set.uuid, job.uuid,
                None, result, zuul_event_id=build.zuul_event_id)
            self.result_events[manager.tenant.name][manager.pipeline.name].put(
                completed_event
            )

            return

        # Update zuul attempts after addBuild above to ensure build_set
        # is up to date.
        attempts = build.build_set.getTries(job)
        params["zuul"]['attempts'] = attempts
        params['zuul']['max_attempts'] = job.attempts
        # TODO (swestphahl): Remove deprecated 'max_attempts' parameter
        params['max_attempts'] = job.attempts

        # Store the Node(set)Request ID in the job arguments, so we can look
        # it up on the executor side to lock the nodes.
        if req_id := build.build_set.getJobNodeRequestID(job):
            if self.sched.nodepool.isNodeRequestID(req_id):
                params["noderequest_id"] = req_id
            else:
                # Note: We'll only create nodeset requests when all
                # components are updated to the model API version supporting
                # those requests, so we don't need to check for that here.
                params["nodesetrequest_id"] = req_id

        zone_known = False
        if executor_zone:
            # Check the component registry for executors subscribed to this
            # zone
            for comp in self.sched.component_registry.all(kind="executor"):
                if comp.zone == executor_zone:
                    zone_known = True
                    break

            if not zone_known:
                self.log.warning(
                    "Job requested '%s' zuul-executor zone, but no "
                    "zuul-executors found for this zone; ignoring zone "
                    "request", executor_zone)
                # Fall back to the default zone
                executor_zone = None

        with trace.use_span(build_span):
            request = BuildRequest(
                uuid=uuid,
                build_set_uuid=build.build_set.uuid,
                job_uuid=job.uuid,
                tenant_name=manager.tenant.name,
                pipeline_name=manager.pipeline.name,
                zone=executor_zone,
                event_id=item.event.zuul_event_id,
                precedence=PRIORITY_MAP[pipeline.precedence],
            )
        self.executor_api.submit(request, params)
        build.updateAttributes(manager.current_context,
                               build_request_ref=request.path)

    def cancel(self, build):
        log = get_annotated_logger(self.log, build.zuul_event_id,
                                   build=build.uuid)
        # Returns whether a running build was canceled
        log.info("Cancel build %s for job %s", build, build.job)

        build.updateAttributes(
            build.build_set.item.manager.current_context,
            canceled=True)

        if not build.build_request_ref:
            log.debug("Build has not been submitted to ZooKeeper")
            return False

        build_request = self.executor_api.get(build.build_request_ref)
        if build_request:
            log.debug("Canceling build request %s", build_request)
            # If we can acquire the build request lock here, the build wasn't
            # picked up by any executor server yet. With acquiring the lock
            # we prevent the executor server from picking up the build so we
            # can cancel it before it will run.
            if self.executor_api.lock(build_request, blocking=False):
                # Mark the build request as complete and forward the event to
                # the scheduler, so the executor server doesn't pick up the
                # request. The build will be deleted from the scheduler when it
                # picks up the BuildCompletedEvent.
                try:
                    if build_request.state == BuildRequest.COMPLETED:
                        log.debug("Build request already completed; skipping")
                        return False
                    log.debug(
                        "Canceling build %s directly because it is not locked"
                        "by any executor", build_request,
                    )
                    build_request.state = BuildRequest.COMPLETED
                    self.executor_api.update(build_request)

                    result = {"result": "CANCELED", "end_time": time.time()}
                    tenant_name = build.build_set.item.manager.tenant.name
                    pipeline_name = build.build_set.item.manager.pipeline.name
                    event = BuildCompletedEvent(
                        build_request.uuid, build_request.build_set_uuid,
                        build_request.job_uuid,
                        build_request.path, result)
                    self.result_events[tenant_name][pipeline_name].put(event)
                finally:
                    self.executor_api.unlock(build_request)
            else:
                log.debug(
                    "Sending cancel request for build %s because it is locked",
                    build_request,
                )
                # If the build request is locked, schedule a cancel request in
                # the executor server.
                self.executor_api.requestCancel(build_request)

            log.debug("Canceled build")
            return True

        return False

    def resumeBuild(self, build):
        log = get_annotated_logger(self.log, build.zuul_event_id)

        if not build.build_request_ref:
            log.debug("Build has not been submitted")
            return False

        build_request = self.executor_api.get(build.build_request_ref)
        if build_request:
            log.debug("Requesting resume for build %s", build)
            self.executor_api.requestResume(build_request)
            return True
        return False

    def removeBuild(self, build):
        log = get_annotated_logger(self.log, build.zuul_event_id)
        log.debug("Removing build %s", build.uuid)

        if not build.build_request_ref:
            log.debug("Build %s has not been submitted to ZooKeeper",
                      build.uuid)
            return

        build_request = self.executor_api.get(build.build_request_ref)
        if build_request:
            # TODO (felix): We could directly remove the build request via
            # its path in ZK to spare a read operation. Usually there should
            # be no need to look up the build request object from ZooKeeper
            # just to immediately remove it.
            self.executor_api.remove(build_request)

    def cleanupLostBuildRequests(self):
        for build_request in self.executor_api.lostRequests():
            try:
                self.cleanupLostBuildRequest(build_request)
            except Exception:
                self.log.exception("Exception cleaning up lost build request:")

    def cleanupLostBuildRequest(self, build_request):
        result = {"result": "ABORTED"}

        # TODO (felix): Once the builds are stored in ZooKeeper, we can store
        # the end_time directly on the build. But for now we have to use the
        # result dict for that.
        result["end_time"] = time.time()

        build_request.state = BuildRequest.COMPLETED
        try:
            self.executor_api.update(build_request)
        except JobRequestNotFound as e:
            self.log.warning("Could not complete build: %s", str(e))
            # In case we re-created the lock directory, still remove
            # the request for the side effect of removing the lock.
            self.executor_api.remove(build_request)
            return
        except BadVersionError:
            # There could be a race condition:
            # The build is found by lost_builds in state RUNNING
            # but gets completed/unlocked before the is_locked()
            # check. Since we use the znode version, the update
            # will fail in this case and we can simply ignore the
            # exception.
            return

        # No need to unlock the build, as it is by definition unlocked

        event = BuildCompletedEvent(
            build_request.uuid, build_request.build_set_uuid,
            build_request.job_uuid,
            build_request.path, result)
        self.result_events[build_request.tenant_name][
            build_request.pipeline_name].put(event)



# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import multiprocessing
import os

import nox


nox.options.error_on_external_run = True
nox.options.reuse_existing_virtualenvs = True
nox.options.sessions = ["tests-3", "linters"]


def set_env(session, var, default):
    session.env[var] = os.environ.get(var, default)


def set_standard_env_vars(session):
    set_env(session, 'OS_LOG_CAPTURE', '1')
    set_env(session, 'OS_STDERR_CAPTURE', '1')
    set_env(session, 'OS_STDOUT_CAPTURE', '1')
    set_env(session, 'OS_TEST_TIMEOUT', '360')
    session.env['PYTHONWARNINGS'] = ','.join([
        'always::DeprecationWarning:zuul.driver.sql.sqlconnection',
        'always::DeprecationWarning:tests.base',
        'always::DeprecationWarning:tests.unit.test_database',
        'always::DeprecationWarning:zuul.driver.sql.alembic.env',
        'always::DeprecationWarning:zuul.driver.sql.alembic.script',
    ])
    # Set PYTHONTRACEMALLOC to a value greater than 0 in the calling env
    # to get tracebacks of that depth for ResourceWarnings. Disabled by
    # default as this consumes more resources and is slow.
    set_env(session, 'PYTHONTRACEMALLOC', '0')


@nox.session(python='3')
def bindep(session):
    set_standard_env_vars(session)
    session.install('bindep')
    session.run('bindep', 'test')


@nox.session(python='3')
def cover(session):
    set_standard_env_vars(session)
    session.env['PYTHON'] = 'coverage run --source zuul --parallel-mode'
    session.install('-r', 'requirements.txt',
                    '-r', 'test-requirements.txt')
    session.install('-e', '.')
    session.run('stestr', 'run')
    session.run('coverage', 'combine')
    session.run('coverage', 'html', '-d', 'cover')
    session.run('coverage', 'xml', '-o', 'cover/coverage.xml')


@nox.session(python='3')
def docs(session):
    set_standard_env_vars(session)
    # Setuptools 82.0.0 removed pkg_resources entirely. sphinxcontrib-openapi
    # depends on pkg_resources and docs builds fail without this pin. We can
    # remove the setuptools pin once sphinxcontrib-openapi drops the use of
    # pkg_resources.
    session.install('setuptools<82.0.0')
    session.install('-r', 'doc/requirements.txt',
                    '-r', 'test-requirements.txt')
    session.install('-e', '.')
    session.run('sphinx-build', '-E', '-W', '-d', 'doc/build/doctrees',
                '-b', 'html', 'doc/source/', 'doc/build/html')


@nox.session(python='3')
def linters(session):
    set_standard_env_vars(session)
    session.install('flake8', 'openapi-spec-validator')
    session.run('flake8')
    session.run('openapi-spec-validator', 'web/public/openapi.yaml')


@nox.session(python='3')
def tests(session):
    set_standard_env_vars(session)
    session.install('-r', 'requirements.txt',
                    '-r', 'test-requirements.txt')
    session.install('-e', '.')
    session.run_always('tools/yarn-build.sh', external=True)
    session.run_always('zuul-manage-ansible', '-v')
    procs = max(int(multiprocessing.cpu_count() * 0.7), 1)
    session.run('stestr', 'run', '--slowest', f'--concurrency={procs}',
                *session.posargs)


@nox.session(python='3')
def upgrade(session):
    set_standard_env_vars(session)
    session.install('-r', 'requirements.txt',
                    '-r', 'test-requirements.txt')
    session.install('-e', '.')
    session.run_always('zuul-manage-ansible', '-v')
    procs = max(int(multiprocessing.cpu_count() * 0.75), 1)
    session.run('stestr', 'run', '--test-path', './tests/upgrade',
                '--slowest', f'--concurrency={procs}',
                *session.posargs)
    # Output the test log to stdout so we have a copy of even the
    # successful output.  We capture and output instead of just
    # streaming it so that it's not interleaved.
    session.run('stestr', 'last', '--all-attachments')


@nox.session(python='3')
def remote(session):
    set_standard_env_vars(session)
    session.install('-r', 'requirements.txt',
                    '-r', 'test-requirements.txt')
    session.install('-e', '.')
    session.run_always('zuul-manage-ansible', '-v')
    session.run('stestr', 'run', '--test-path', './tests/remote',
                *session.posargs)


@nox.session(python='3')
def venv(session):
    set_standard_env_vars(session)
    session.install('-r', 'requirements.txt',
                    '-r', 'test-requirements.txt')
    session.install('-e', '.')
    session.run(*session.posargs)


@nox.session(python='3')
def zuul_client(session):
    set_standard_env_vars(session)
    session.install('zuul-client',
                    '-r', 'test-requirements.txt',
                    '-r', 'requirements.txt')
    session.install('-e', '.')
    session.run_always('zuul-manage-ansible', '-v')
    session.run(
        'stestr', 'run', '--concurrency=1',
        '--test-path', './tests/zuul_client')



# Copyright 2021-2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import difflib
import os
import re
import subprocess
import time
import uuid

from zuul.driver.sql import SQLDriver
from zuul.zk import ZooKeeperClient
from tests.base import (
    BaseTestCase,
    MySQLSchemaFixture,
    PostgresqlSchemaFixture,
    ZOOKEEPER_SESSION_TIMEOUT,
)

import sqlalchemy
import testtools


class DBBaseTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.setupZK()

        self.zk_client = ZooKeeperClient(
            self.zk_chroot_fixture.zk_hosts,
            tls_cert=self.zk_chroot_fixture.zookeeper_cert,
            tls_key=self.zk_chroot_fixture.zookeeper_key,
            tls_ca=self.zk_chroot_fixture.zookeeper_ca,
            timeout=ZOOKEEPER_SESSION_TIMEOUT,
        )
        self.addCleanup(self.zk_client.disconnect)
        self.zk_client.connect()


class CommonDBTestMixin:
    def test_system_events(self):
        tenant = 'tenant1',
        event_id = 'deadbeef'
        event_time = datetime.datetime.fromtimestamp(
            1763581571, tz=datetime.timezone.utc)
        event_type = "reconfiguration"
        description = "A reconfiguration happened"

        # Create using the driver-external interface
        with self.connection.getSession() as db:
            db.createSystemEvent(
                tenant=tenant,
                event_id=event_id,
                event_time=event_time,
                event_type=event_type,
                description=description,
            )

        # Verify that worked using the driver-external interface
        self.assertEqual(len(self.connection.getSystemEvents()), 1)
        self.assertEqual(self.connection.getSystemEvents()[0].event_id,
                         event_id)

        now = time.time()
        # Create a newer event
        event_time = datetime.datetime.fromtimestamp(
            now, tz=datetime.timezone.utc)
        with self.connection.getSession() as db:
            db.createSystemEvent(
                tenant=tenant,
                event_id=event_id,
                event_time=event_time,
                event_type=event_type,
                description=description,
            )
        self.assertEqual(len(self.connection.getSystemEvents()), 2)

        # Delete the old event
        cutoff = datetime.datetime.fromtimestamp(
            now - 10, tz=datetime.timezone.utc)
        self.connection.deleteSystemEvents(cutoff)

        self.assertEqual(len(self.connection.getSystemEvents()), 1)


class TestMysqlDatabase(DBBaseTestCase, CommonDBTestMixin):
    def setUp(self):
        super().setUp()

        f = MySQLSchemaFixture(self.id(), self.random_databases,
                               self.delete_databases)
        self.useFixture(f)

        config = dict(dburi=f.dburi)
        driver = SQLDriver()
        self.connection = driver.getConnection('database', config)
        self.connection.onLoad(self.zk_client)
        self.addCleanup(self._cleanup)

    def _cleanup(self):
        self.connection.onStop()

    def compareMysql(self, alembic_text, sqlalchemy_text):
        alembic_lines = alembic_text.split('\n')
        sqlalchemy_lines = sqlalchemy_text.split('\n')
        self.assertEqual(len(alembic_lines), len(sqlalchemy_lines))
        alembic_constraints = []
        sqlalchemy_constraints = []
        for i in range(len(alembic_lines)):
            if alembic_lines[i].startswith("  `"):
                # Column
                self.assertEqual(alembic_lines[i], sqlalchemy_lines[i])
            elif alembic_lines[i].startswith("  "):
                # Constraints can be unordered
                # strip trailing commas since the last line omits it
                alembic_constraints.append(
                    re.sub(',$', '', alembic_lines[i]))
                sqlalchemy_constraints.append(
                    re.sub(',$', '', sqlalchemy_lines[i]))
            else:
                self.assertEqual(alembic_lines[i], sqlalchemy_lines[i])
        alembic_constraints.sort()
        sqlalchemy_constraints.sort()
        self.assertEqual(alembic_constraints, sqlalchemy_constraints)

    def test_migration(self):
        # Test that SQLAlchemy create_all produces the same output as
        # a full migration run.
        sqlalchemy_tables = {}
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("set foreign_key_checks=0")
            for table in connection.exec_driver_sql("show tables"):
                table = table[0]
                sqlalchemy_tables[table] = connection.exec_driver_sql(
                    f"show create table {table}").one()[1]
                connection.exec_driver_sql(f"drop table {table}")
            connection.exec_driver_sql("set foreign_key_checks=1")

        self.connection.force_migrations = True
        self.connection.onLoad(self.zk_client)
        with self.connection.engine.begin() as connection:
            for table in connection.exec_driver_sql("show tables"):
                table = table[0]
                create = connection.exec_driver_sql(
                    f"show create table {table}").one()[1]
                self.compareMysql(create, sqlalchemy_tables[table])

    def test_migration_4647def24b32(self):
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("set foreign_key_checks=0")
            for table in connection.exec_driver_sql("show tables"):
                table = table[0]
                connection.exec_driver_sql(f"drop table {table}")
            connection.exec_driver_sql("set foreign_key_checks=1")

        self.connection.force_migrations = True
        self.connection._migrate('c57e9e76b812')
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql(
                "insert into zuul_buildset (project, result) "
                "values ('org/project', 'SUCCESS')")
            connection.exec_driver_sql(
                "insert into zuul_buildset (project, result) "
                "values ('org/project', 'MERGER_FAILURE')")
            results = [r[0] for r in connection.exec_driver_sql(
                "select result from zuul_buildset")]
            self.assertEqual(results, ['SUCCESS', 'MERGER_FAILURE'])

        self.connection._migrate()
        with self.connection.engine.begin() as connection:
            results = [r[0] for r in connection.exec_driver_sql(
                "select result from zuul_buildset")]
            self.assertEqual(results, ['SUCCESS', 'MERGE_CONFLICT'])

    def test_migration_c7467b642498(self):
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("set foreign_key_checks=0")
            for table in connection.exec_driver_sql("show tables"):
                table = table[0]
                connection.exec_driver_sql(f"drop table {table}")
            connection.exec_driver_sql("set foreign_key_checks=1")

        self.connection.force_migrations = True
        self.connection._migrate('4647def24b32')
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql(
                "insert into zuul_buildset (project, result) "
                "values ('org/project', 'SUCCESS')")
            connection.exec_driver_sql(
                "insert into zuul_buildset "
                "(project, result, first_build_start_time) "
                "values ('org/project', 'SUCCESS', '2022-05-01 12:34:56')")
            connection.exec_driver_sql(
                "insert into zuul_buildset "
                "(project, result, last_build_end_time) "
                "values ('org/project', 'SUCCESS', '2022-05-02 12:34:56')")
            connection.exec_driver_sql(
                "insert into zuul_buildset "
                "(project, result, event_timestamp) "
                "values ('org/project', 'SUCCESS', '2022-05-03 12:34:56')")
            connection.exec_driver_sql(
                "insert into zuul_buildset (project, result, "
                "first_build_start_time, "
                "last_build_end_time, "
                "event_timestamp)"
                "values ('org/project', 'SUCCESS', "
                "'2022-05-11 12:34:56', "
                "'2022-05-12 12:34:56', "
                "'2022-05-13 12:34:56')")

        self.connection._migrate()
        with self.connection.engine.begin() as connection:
            results = [str(r[0]) for r in connection.exec_driver_sql(
                "select updated from zuul_buildset")]
            self.assertEqual(results,
                             ['1970-01-01 00:00:00',
                              '2022-05-01 12:34:56',
                              '2022-05-02 12:34:56',
                              '2022-05-03 12:34:56',
                              '2022-05-13 12:34:56'])

    def test_migration_f7843ddf1552(self):
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("set foreign_key_checks=0")
            for table in connection.exec_driver_sql("show tables"):
                table = table[0]
                connection.exec_driver_sql(f"drop table {table}")
            connection.exec_driver_sql("set foreign_key_checks=1")

        self.connection.force_migrations = True
        self.connection._migrate('151893067f91')
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("""
            insert into zuul_buildset
            (id, uuid, zuul_ref, project, `change`, patchset, ref,
             ref_url, oldrev, newrev, branch)
            values
            (1, "bsuuid1", "Z1", "project1",
            1, "1",
            "refs/changes/1",
             "http://project1/1",
            "2d48fe5afe0bc7785294b28e9e96a6c622945b9d",
            "8a938aff20d691d1b9a9f461c6375f0c45acd305",
            "master"),
            (2, "bsuuid2", "Z2", "project1",
            2, "ee743613ce5b3aee11d12e91e932d7876bc0b40c",
            "refs/changes/2",
             "http://project1/2",
            "bd11c4ff79245ae555418e19840ce4752e48ea38",
            "d1ffc204e4b4955f2a47e7e37618fcf13bb1b9fd",
            "stable"),
            (3, "bsuuid3", "Z3", "project2", NULL, NULL, "refs/tags/foo",
             "http://project3",
            "a6cf0e07e9e2964a810196f764fc4ac766742568",
            "1a226acbd022968914574fdeb467b78bc5bfcc77",
            NULL)
            """)
            connection.exec_driver_sql("""
            insert into zuul_build
            (id, buildset_id, uuid, job_name, result)
            values
            (1, 1, "builduuid1", "job1", "RESULT1"),
            (2, 1, "builduuid2", "job2", "RESULT2"),
            (3, 2, "builduuid3", "job1", "RESULT3"),
            (4, 2, "builduuid4", "job2", "RESULT4"),
            (5, 3, "builduuid5", "job3", "RESULT5"),
            (6, 3, "builduuid6", "job4", "RESULT6")
            """)

        self.connection._migrate()
        with self.connection.engine.begin() as connection:
            results = [r for r in connection.exec_driver_sql(
                "select b.id, r.ref from zuul_build b join zuul_ref r "
                "on b.ref_id=r.id order by b.id")]
            self.assertEqual(results, [
                (1, 'refs/changes/1'),
                (2, 'refs/changes/1'),
                (3, 'refs/changes/2'),
                (4, 'refs/changes/2'),
                (5, 'refs/tags/foo'),
                (6, 'refs/tags/foo'),
            ])
            results = [r for r in connection.exec_driver_sql(
                "select bs.uuid, r.ref "
                "from zuul_buildset bs, zuul_buildset_ref t, zuul_ref r "
                "where bs.id = t.buildset_id and r.id = t.ref_id "
                "order by bs.uuid")]
            self.assertEqual(results, [
                ('bsuuid1', 'refs/changes/1'),
                ('bsuuid2', 'refs/changes/2'),
                ('bsuuid3', 'refs/tags/foo'),
            ])

    def test_migration_f7843ddf1552_failure(self):
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("set foreign_key_checks=0")
            for table in connection.exec_driver_sql("show tables"):
                table = table[0]
                connection.exec_driver_sql(f"drop table {table}")
            connection.exec_driver_sql("set foreign_key_checks=1")

        self.connection.force_migrations = True
        self.connection._migrate('151893067f91')
        # This test is identical to the one above, except the patchset
        # sha for the first buildset row is too long to fit in the
        # column (it's lengith is double).  The insert here is fine,
        # but that will cause the migration to fail.
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("""
            insert into zuul_buildset
            (id, uuid, zuul_ref, project, `change`, patchset, ref,
             ref_url, oldrev, newrev, branch)
            values
            (1, "bsuuid1", "Z1", "project1",
            1, "1",
            "refs/changes/1",
             "http://project1/1",
            "2d48fe5afe0bc7785294b28e9e96a6c622945b9d"
            "2d48fe5afe0bc7785294b28e9e96a6c622945b9d",
            "8a938aff20d691d1b9a9f461c6375f0c45acd305",
            "master"),
            (2, "bsuuid2", "Z2", "project1",
            2, "ee743613ce5b3aee11d12e91e932d7876bc0b40c",
            "refs/changes/2",
             "http://project1/2",
            "bd11c4ff79245ae555418e19840ce4752e48ea38",
            "d1ffc204e4b4955f2a47e7e37618fcf13bb1b9fd",
            "stable"),
            (3, "bsuuid3", "Z3", "project2", NULL, NULL, "refs/tags/foo",
             "http://project3",
            "a6cf0e07e9e2964a810196f764fc4ac766742568",
            "1a226acbd022968914574fdeb467b78bc5bfcc77",
            NULL)
            """)
            connection.exec_driver_sql("""
            insert into zuul_build
            (id, buildset_id, uuid, job_name, result)
            values
            (1, 1, "builduuid1", "job1", "RESULT1"),
            (2, 1, "builduuid2", "job2", "RESULT2"),
            (3, 2, "builduuid3", "job1", "RESULT3"),
            (4, 2, "builduuid4", "job2", "RESULT4"),
            (5, 3, "builduuid5", "job3", "RESULT5"),
            (6, 3, "builduuid6", "job4", "RESULT6")
            """)

        with testtools.ExpectedException(sqlalchemy.exc.DataError):
            self.connection._migrate()

        with self.connection.engine.begin() as connection:
            tables = [r[0] for r in connection.exec_driver_sql(
                "show tables")]
            # Make sure we rolled back the buildset_new table creation
            self.assertNotIn('zuul_buildset_new', tables)
            self.assertIn('zuul_buildset', tables)

    def test_migration_d0a269345a22(self):
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("set foreign_key_checks=0")
            for table in connection.exec_driver_sql("show tables"):
                table = table[0]
                connection.exec_driver_sql(f"drop table {table}")
            connection.exec_driver_sql("set foreign_key_checks=1")

        self.connection.force_migrations = True
        self.connection._migrate('f7843ddf1552')
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("""
            insert into zuul_ref
            (id, project, ref, ref_url, `change`,
            patchset, oldrev, newrev, branch)
            values
            (473951, 'openstack/neutron', 'refs/heads/master',
            'https://git.openstack.org/cgit/openstack/neutron/commit/?id=None',
            0, '', '', '', ''),
            (473958, 'openstack/neutron', 'refs/heads/master',
            'https://git.openstack.org/cgit/openstack/neutron/commit/'
            '?id=fd024ac468456561918e02d32f55eabada102fd3',
            0, '',
            '0017b625ddb64cb3f383e1a4b7f20b7a62153705',
            'fd024ac468456561918e02d32f55eabada102fd3',
            ''),
            (481941, 'openstack/neutron', 'refs/heads/stable/zed',
            'https://opendev.org/openstack/neutron/commit/'
            'd6ee668cc32725cb7d15d2e08fdb50a761f91fe4',
            0, '',
            'fdba20b27a1b97459db1e83f9b235149fe7fd1f7',
            'd6ee668cc32725cb7d15d2e08fdb50a761f91fe4',
            'stable/zed');
            """)
        self.connection._migrate()
        with self.connection.engine.begin() as connection:
            results = [r for r in connection.exec_driver_sql(
                "select id, ref, branch from zuul_ref order by id")]
            self.assertEqual(results, [
                (473951, 'refs/heads/master', 'master'),
                (473958, 'refs/heads/master', 'master'),
                (481941, 'refs/heads/stable/zed', 'stable/zed'),
            ])

    def test_buildsets(self):
        tenant = 'tenant1',
        buildset_uuid = 'deadbeef'
        change = 1234

        # Create the buildset entry (driver-internal interface)
        with self.connection.getSession() as db:
            bs = db.createBuildSet(
                uuid=buildset_uuid,
                tenant=tenant,
                pipeline='check',
                event_id='eventid',
            )
            ref = db.getOrCreateRef(
                project='project',
                change=change,
                patchset='1',
                ref='',
                oldrev='',
                newrev='',
                branch='master',
                ref_url='http://example.com/1234',
            )
            bs.refs.append(ref)

        # Verify that worked using the driver-external interface
        self.assertEqual(len(self.connection.getBuildsets()), 1)
        self.assertEqual(self.connection.getBuildsets()[0].uuid, buildset_uuid)

        # Update the buildset using the internal interface
        with self.connection.getSession() as db:
            db_buildset = db.getBuildset(tenant=tenant, uuid=buildset_uuid)
            self.assertEqual(db_buildset.refs[0].change, change)
            db_buildset.result = 'SUCCESS'

        # Verify that worked
        db_buildset = self.connection.getBuildset(
            tenant=tenant, uuid=buildset_uuid)
        self.assertEqual(db_buildset.result, 'SUCCESS')

    def test_sanitize_subtring_query(self):
        session = self.connection.getSession()
        self.assertEqual(None, session._sanitizeSubstringQuery(None))
        self.assertEqual(
            "job-name", session._sanitizeSubstringQuery("job-name"))
        self.assertEqual(
            "$%job-$%name$%", session._sanitizeSubstringQuery("%job-%name%"))
        self.assertEqual(
            "job$_-$%name$%", session._sanitizeSubstringQuery("job_-%name%"))
        self.assertEqual(
            "job$_-$%name$%", session._sanitizeSubstringQuery("job_-%name%"))
        self.assertEqual(
            "$$$%job$_-$%name$%",
            session._sanitizeSubstringQuery("$%job_-%name%"))

    def test_fuzzy_filter_op(self):
        col = sqlalchemy.Column("job_name")
        session = self.connection.getSession()

        filter = session._getFuzzyFilterOp(col, "job-name")
        self.assertEqual(type(col == "foo"), type(filter))
        self.assertEqual("job-name", filter.right.value)

        filter = session._getFuzzyFilterOp(col, "job%name")
        self.assertEqual(type(col == "foo"), type(filter))
        self.assertEqual("job%name", filter.right.value)

        filter = session._getFuzzyFilterOp(col, "job$name")
        self.assertEqual(type(col == "foo"), type(filter))
        self.assertEqual("job$name", filter.right.value)

        filter = session._getFuzzyFilterOp(col, "job_name")
        self.assertEqual(type(col == "foo"), type(filter))
        self.assertEqual("job_name", filter.right.value)

        filter = session._getFuzzyFilterOp(col, "*job*name*")
        self.assertEqual(type(col.like("foo")), type(filter))
        self.assertEqual("$", filter.modifiers["escape"])
        self.assertEqual("%job%name%", filter.right.value)

        filter = session._getFuzzyFilterOp(col, None)
        self.assertEqual(type(col.__eq__(None)), type(filter))
        self.assertEqual(sqlalchemy.sql.elements.Null, type(filter.right))


class TestPostgresqlDatabase(DBBaseTestCase, CommonDBTestMixin):
    def setUp(self):
        super().setUp()

        f = PostgresqlSchemaFixture(self.id(), self.random_databases,
                                    self.delete_databases)
        self.useFixture(f)
        self.db = f

        config = dict(dburi=f.dburi)
        driver = SQLDriver()
        self.connection = driver.getConnection('database', config)
        self.connection.onLoad(self.zk_client)
        self.addCleanup(self._cleanup)

    def _cleanup(self):
        self.connection.onStop()

    def test_migration(self):
        # Test that SQLAlchemy create_all produces the same output as
        # a full migration run.
        pg_dump = os.environ.get("ZUUL_TEST_PG_DUMP", "pg_dump")
        # Force a restrict key so we can compare the output
        # of the dump before and after the migration.
        restrict_key = uuid.uuid4().hex
        sqlalchemy_out = subprocess.check_output(
            f"{pg_dump} -h {self.db.host} -U {self.db.name} -s {self.db.name}"
            f" --restrict-key={restrict_key}",
            shell=True,
            env={'PGPASSWORD': self.db.passwd}
        )

        with self.connection.engine.begin() as connection:
            tables = [x[0] for x in connection.exec_driver_sql(
                "select tablename from pg_catalog.pg_tables "
                "where schemaname='public'"
            ).all()]

            self.assertTrue(len(tables) > 0)
            for table in tables:
                connection.exec_driver_sql(f"drop table {table} cascade")

        self.connection.force_migrations = True
        self.connection.onLoad(self.zk_client)

        alembic_out = subprocess.check_output(
            f"{pg_dump} -h {self.db.host} -U {self.db.name} -s {self.db.name}"
            f" --restrict-key={restrict_key}",
            shell=True,
            env={'PGPASSWORD': self.db.passwd}
        )
        try:
            self.assertEqual(alembic_out, sqlalchemy_out)
        except Exception:
            differ = difflib.Differ()
            alembic_out = alembic_out.decode('utf8').splitlines()
            sqlalchemy_out = sqlalchemy_out.decode('utf8').splitlines()
            diff = '\n'.join(list(differ.compare(alembic_out, sqlalchemy_out)))
            self.log.debug("Diff:\n%s", diff)
            raise

    def test_migration_f7843ddf1552(self):
        with self.connection.engine.begin() as connection:
            tables = [x[0] for x in connection.exec_driver_sql(
                "select tablename from pg_catalog.pg_tables "
                "where schemaname='public'"
            ).all()]

            self.assertTrue(len(tables) > 0)
            for table in tables:
                connection.exec_driver_sql(f"drop table {table} cascade")

        self.connection.force_migrations = True
        self.connection._migrate('151893067f91')
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("""
            insert into zuul_buildset
            (id, uuid, zuul_ref, project, change, patchset, ref,
             ref_url, oldrev, newrev, branch)
            values
            (1, 'bsuuid1', 'Z1', 'project1',
            1, '1',
            'refs/changes/1',
             'http://project1/1',
            '2d48fe5afe0bc7785294b28e9e96a6c622945b9d',
            '8a938aff20d691d1b9a9f461c6375f0c45acd305',
            'master'),
            (2, 'bsuuid2', 'Z2', 'project1',
            2, 'ee743613ce5b3aee11d12e91e932d7876bc0b40c',
            'refs/changes/2',
             'http://project1/2',
            'bd11c4ff79245ae555418e19840ce4752e48ea38',
            'd1ffc204e4b4955f2a47e7e37618fcf13bb1b9fd',
            'stable'),
            (3, 'bsuuid3', 'Z3', 'project2', NULL, NULL, 'refs/tags/foo',
             'http://project3',
            'a6cf0e07e9e2964a810196f764fc4ac766742568',
            '1a226acbd022968914574fdeb467b78bc5bfcc77',
            NULL)
            """)
            connection.exec_driver_sql("""
            insert into zuul_build
            (id, buildset_id, uuid, job_name, result)
            values
            (1, 1, 'builduuid1', 'job1', 'RESULT1'),
            (2, 1, 'builduuid2', 'job2', 'RESULT2'),
            (3, 2, 'builduuid3', 'job1', 'RESULT3'),
            (4, 2, 'builduuid4', 'job2', 'RESULT4'),
            (5, 3, 'builduuid5', 'job3', 'RESULT5'),
            (6, 3, 'builduuid6', 'job4', 'RESULT6')
            """)

        self.connection._migrate()
        with self.connection.engine.begin() as connection:
            results = [r for r in connection.exec_driver_sql(
                "select b.id, r.ref from zuul_build b join zuul_ref r "
                "on b.ref_id=r.id order by b.id")]
            self.assertEqual(results, [
                (1, 'refs/changes/1'),
                (2, 'refs/changes/1'),
                (3, 'refs/changes/2'),
                (4, 'refs/changes/2'),
                (5, 'refs/tags/foo'),
                (6, 'refs/tags/foo'),
            ])
            results = [r for r in connection.exec_driver_sql(
                "select bs.uuid, r.ref "
                "from zuul_buildset bs, zuul_buildset_ref t, zuul_ref r "
                "where bs.id = t.buildset_id and r.id = t.ref_id "
                "order by bs.uuid")]
            self.assertEqual(results, [
                ('bsuuid1', 'refs/changes/1'),
                ('bsuuid2', 'refs/changes/2'),
                ('bsuuid3', 'refs/tags/foo'),
            ])

    def test_migration_f7843ddf1552_failure(self):
        with self.connection.engine.begin() as connection:
            tables = [x[0] for x in connection.exec_driver_sql(
                "select tablename from pg_catalog.pg_tables "
                "where schemaname='public'"
            ).all()]

            self.assertTrue(len(tables) > 0)
            for table in tables:
                connection.exec_driver_sql(f"drop table {table} cascade")

        self.connection.force_migrations = True
        self.connection._migrate('151893067f91')
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("""
            insert into zuul_buildset
            (id, uuid, zuul_ref, project, change, patchset, ref,
             ref_url, oldrev, newrev, branch)
            values
            (1, 'bsuuid1', 'Z1', 'project1',
            1, '1',
            'refs/changes/1',
             'http://project1/1',
            '2d48fe5afe0bc7785294b28e9e96a6c622945b9d'
            '2d48fe5afe0bc7785294b28e9e96a6c622945b9d',
            '8a938aff20d691d1b9a9f461c6375f0c45acd305',
            'master'),
            (2, 'bsuuid2', 'Z2', 'project1',
            2, 'ee743613ce5b3aee11d12e91e932d7876bc0b40c',
            'refs/changes/2',
             'http://project1/2',
            'bd11c4ff79245ae555418e19840ce4752e48ea38',
            'd1ffc204e4b4955f2a47e7e37618fcf13bb1b9fd',
            'stable'),
            (3, 'bsuuid3', 'Z3', 'project2', NULL, NULL, 'refs/tags/foo',
             'http://project3',
            'a6cf0e07e9e2964a810196f764fc4ac766742568',
            '1a226acbd022968914574fdeb467b78bc5bfcc77',
            NULL)
            """)
            connection.exec_driver_sql("""
            insert into zuul_build
            (id, buildset_id, uuid, job_name, result)
            values
            (1, 1, 'builduuid1', 'job1', 'RESULT1'),
            (2, 1, 'builduuid2', 'job2', 'RESULT2'),
            (3, 2, 'builduuid3', 'job1', 'RESULT3'),
            (4, 2, 'builduuid4', 'job2', 'RESULT4'),
            (5, 3, 'builduuid5', 'job3', 'RESULT5'),
            (6, 3, 'builduuid6', 'job4', 'RESULT6')
            """)

        with testtools.ExpectedException(sqlalchemy.exc.DataError):
            self.connection._migrate()

        with self.connection.engine.begin() as connection:
            tables = [x[0] for x in connection.exec_driver_sql(
                "select tablename from pg_catalog.pg_tables "
                "where schemaname='public'"
            ).all()]
            # Make sure we rolled back the buildset_new table creation
            self.assertNotIn('zuul_buildset_new', tables)
            self.assertIn('zuul_buildset', tables)

    def test_migration_d0a269345a22(self):
        with self.connection.engine.begin() as connection:
            tables = [x[0] for x in connection.exec_driver_sql(
                "select tablename from pg_catalog.pg_tables "
                "where schemaname='public'"
            ).all()]

            self.assertTrue(len(tables) > 0)
            for table in tables:
                connection.exec_driver_sql(f"drop table {table} cascade")

        self.connection.force_migrations = True
        self.connection._migrate('f7843ddf1552')
        with self.connection.engine.begin() as connection:
            connection.exec_driver_sql("""
            insert into zuul_ref
            (id, project, ref, ref_url, change,
            patchset, oldrev, newrev, branch)
            values
            (473951, 'openstack/neutron', 'refs/heads/master',
            'https://git.openstack.org/cgit/openstack/neutron/commit/?id=None',
            0, '', '', '', ''),
            (473958, 'openstack/neutron', 'refs/heads/master',
            'https://git.openstack.org/cgit/openstack/neutron/commit/'
            '?id=fd024ac468456561918e02d32f55eabada102fd3',
            0, '',
            '0017b625ddb64cb3f383e1a4b7f20b7a62153705',
            'fd024ac468456561918e02d32f55eabada102fd3',
            ''),
            (481941, 'openstack/neutron', 'refs/heads/stable/zed',
            'https://opendev.org/openstack/neutron/commit/'
            'd6ee668cc32725cb7d15d2e08fdb50a761f91fe4',
            0, '',
            'fdba20b27a1b97459db1e83f9b235149fe7fd1f7',
            'd6ee668cc32725cb7d15d2e08fdb50a761f91fe4',
            'stable/zed');
            """)
        self.connection._migrate()
        with self.connection.engine.begin() as connection:
            results = [r for r in connection.exec_driver_sql(
                "select id, ref, branch from zuul_ref order by id")]
            self.assertEqual(results, [
                (473951, 'refs/heads/master', 'master'),
                (473958, 'refs/heads/master', 'master'),
                (481941, 'refs/heads/stable/zed', 'stable/zed'),
            ])



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

from zuul.zk.launcher import LockableZKObjectCache
from zuul.model import ImageBuildArtifact, ImageUpload


class ImageBuildRegistry(LockableZKObjectCache):

    def __init__(self, zk_client, updated_event=None):
        self.builds_by_image_name = collections.defaultdict(set)
        super().__init__(
            zk_client,
            updated_event,
            root=ImageBuildArtifact.ROOT,
            items_path=ImageBuildArtifact.IMAGES_PATH,
            locks_path=ImageBuildArtifact.LOCKS_PATH,
            zkobject_class=ImageBuildArtifact,
        )

    def postCacheHook(self, event, data, stat, key, obj):
        exists = key in self._cached_objects
        if exists:
            if obj:
                builds = self.builds_by_image_name[obj.canonical_name]
                builds.add(key)
        else:
            if obj:
                builds = self.builds_by_image_name[obj.canonical_name]
                builds.discard(key)
        super().postCacheHook(event, data, stat, key, obj)

    def getArtifactsForImage(self, image_canonical_name):
        keys = list(self.builds_by_image_name[image_canonical_name])
        arts = [self._cached_objects.get(key) for key in keys]
        arts = [a for a in arts if a is not None]
        # Sort in a stable order, primarily by timestamp, then format
        # for identical timestamps.
        arts = sorted(arts, key=lambda x: x.format)
        arts = sorted(arts, key=lambda x: x.timestamp)
        return arts

    def getAllArtifacts(self):
        keys = []
        for image_canonical_name in self.builds_by_image_name.keys():
            keys.extend(self.builds_by_image_name[image_canonical_name])
        arts = [self._cached_objects.get(key) for key in keys]
        arts = [a for a in arts if a is not None]
        # Sort in a stable order, primarily by timestamp, then format
        # for identical timestamps.
        arts = sorted(arts, key=lambda x: x.format)
        arts = sorted(arts, key=lambda x: x.timestamp)
        return arts


class ImageUploadRegistry(LockableZKObjectCache):

    def __init__(self, zk_client, updated_event=None):
        self.uploads_by_image_name = collections.defaultdict(set)
        super().__init__(
            zk_client,
            updated_event,
            root=ImageUpload.ROOT,
            items_path=ImageUpload.UPLOADS_PATH,
            locks_path=ImageUpload.LOCKS_PATH,
            zkobject_class=ImageUpload,
        )

    def postCacheHook(self, event, data, stat, key, obj):
        exists = key in self._cached_objects
        if exists:
            if obj:
                uploads = self.uploads_by_image_name[obj.canonical_name]
                uploads.add(key)
        else:
            if obj:
                uploads = self.uploads_by_image_name[obj.canonical_name]
                uploads.discard(key)
        super().postCacheHook(event, data, stat, key, obj)

    def getUploadsForImage(self, image_canonical_name):
        keys = list(self.uploads_by_image_name[image_canonical_name])
        uploads = [self._cached_objects.get(key) for key in keys]
        uploads = [u for u in uploads if u is not None]
        uploads = sorted(uploads, key=lambda x: x.timestamp)
        return uploads

    def getAllUploads(self):
        ret = []
        for image_canonical_name in self.uploads_by_image_name:
            ret.extend(self.getUploadsForImage(image_canonical_name))
        return ret



# -*- coding: utf-8 -*-
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys, os, datetime
import subprocess
import re
from zuul import version

# The minimum version to link to
min_version = (3, 3, 0)

sys.path.insert(0, os.path.abspath('../..'))
# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx_autodoc_typehints',
    'sphinx.ext.graphviz',
    'sphinxcontrib.programoutput',
    'sphinxcontrib.openapi',
    'zuul_sphinx',
    'zuul.sphinx.ansible',
    'reno.sphinxext',
    'sphinx_rtd_theme',
]

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

primary_domain = 'zuul'

# The suffix of source filenames.
source_suffix = {'.rst': 'restructuredtext'}

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Zuul'
copyright = u'2012-%s, Zuul project contributors' % datetime.date.today().year

doc_root = os.environ.get('ZUUL_DOC_ROOT', '/docs/%s' % (project.lower()))

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = "sphinx_rtd_theme"
# This adds a <link rel=canonical> tag
html_baseurl = f'https://zuul-ci.org{doc_root}/latest'

if version.is_release:
    current_version = version.release_string
    version = version.release_string
    versions = [('latest', f'{doc_root}/latest/')]
else:
    # Uncomment this if we want to use the in-development version
    # number (eg 4.10.5.dev4 887cf31e4 )
    # version = version.get_version_string()
    version = 'latest'
    current_version = 'latest'
    versions = [('latest', f'{doc_root}/latest/')]

try:
    output = subprocess.check_output(['git', 'tag']).decode('utf8')
except subprocess.CalledProcessError:
    output = ''

interesting_tags = []
for tag in output.splitlines():
    if re.match(r'^\d+\.\d+\.\d+$', tag):
        parts = tuple(map(int, tag.split('.')))
        if parts < min_version:
            continue
        interesting_tags.append((parts, tag))
for parts, tag in reversed(sorted(interesting_tags, key=lambda x: x[0])):
    versions.append((tag, f'{doc_root}/{tag}/'))

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'collapse_navigation': False,
    'navigation_depth': -1,
    'logo_only': True,
    'style_nav_header_background': '#41B6E6',
}

html_context = {
    # This controls what is displayed at the top of the navbar.
    'version': version,
    # This controls what the caret selection displays at the bottom of
    # the navbar.
    'current_version': current_version,
    # A tuple of (slug, url)
    'versions': versions,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    'custom.css',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/logo.svg'

# Output file base name for HTML help builder.
htmlhelp_basename = '%sdoc' % project

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index',
     '%s.tex' % project,
     u'%s Documentation' % project,
     u'Zuul contributors', 'manual'),
]

# Example configuration for intersphinx: refer to the Python standard library.
#intersphinx_mapping = {'http://docs.python.org/': None}

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Additional Zuul role paths
zuul_role_paths = []



# Copyright (c) 2022 Red Hat
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

# This file, by existing, should be found instead of ansible's built in
# file module.

from ansible.module_utils.basic import *  # noqa
from ansible.module_utils.basic import AnsibleModule


def main():
    module = AnsibleModule(
        argument_spec=dict(
            an_arg=dict(required=True, type='str'),
        )
    )

    results = [
        'this is a string',
        module.params['an_arg'],
        'a final string'
    ]

    module.exit_json(msg="A plugin message", results=results)


if __name__ == '__main__':
    main()



# Copyright 2018-2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import urllib.parse

from zuul.ansible.schema import (
    artifact_schema,
    warning_schema,
)


def validate_schema(data, schema):
    try:
        schema(data)
    except Exception:
        return False
    return True


def get_artifacts_from_result_data(result_data, logger=None):
    ret = []
    if validate_schema(result_data, artifact_schema):
        artifacts = result_data.get('zuul', {}).get(
            'artifacts', [])
        default_url = result_data.get('zuul', {}).get(
            'log_url')
        if default_url:
            if default_url[-1] != '/':
                default_url += '/'
        for artifact in artifacts:
            url = artifact['url']
            if default_url:
                # If the artifact url is relative, it will be combined
                # with the log_url; if it is absolute, it will replace
                # it.
                try:
                    url = urllib.parse.urljoin(default_url, url)
                except Exception:
                    if logger:
                        logger.debug("Error parsing URL:",
                                     exc_info=1)
            d = artifact.copy()
            d['url'] = url
            ret.append(d)
    else:
        if logger:
            logger.debug("Result data did not pass artifact schema "
                         "validation: %s", result_data)
    return ret


def get_warnings_from_result_data(result_data, logger=None):
    if validate_schema(result_data, warning_schema):
        return result_data.get('zuul', {}).get('warnings', [])
    else:
        if logger:
            logger.debug("Result data did not pass warnings schema "
                         "validation: %s", result_data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""merge_conflict_rename

Revision ID: 4647def24b32
Revises: c57e9e76b812
Create Date: 2022-02-24 14:56:52.597907

"""

# revision identifiers, used by Alembic.
revision = '4647def24b32'
down_revision = 'c57e9e76b812'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

BUILDSET_TABLE = 'zuul_buildset'


def upgrade(table_prefix=''):
    buildset = sa.table(table_prefix + BUILDSET_TABLE, sa.column('result'))
    op.execute(buildset.update().
               where(buildset.c.result ==
                     op.inline_literal('MERGER_FAILURE')).
               values({'result': op.inline_literal('MERGE_CONFLICT')}))


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from tests.base import ZuulTestCase, iterate_timeout


class TestTimerTwoTenants(ZuulTestCase):
    tenant_config_file = 'config/timer-two-tenant/main.yaml'

    def test_timer_two_tenants(self):
        # The pipeline triggers every second.  Wait until we have 4
        # jobs (2 from each tenant).
        for _ in iterate_timeout(60, 'jobs started'):
            if len(self.history) >= 4:
                break

        tenant_one_projects = set()
        tenant_two_projects = set()
        for h in self.history:
            if h.parameters['zuul']['tenant'] == 'tenant-one':
                tenant_one_projects.add((h.parameters['items'][0]['project']))
            if h.parameters['zuul']['tenant'] == 'tenant-two':
                tenant_two_projects.add((h.parameters['items'][0]['project']))

        # Verify that the right job ran in the right tenant
        self.assertEqual(tenant_one_projects, {'org/project1'})
        self.assertEqual(tenant_two_projects, {'org/project2'})

        # Stop running timer jobs so the assertions don't race.
        self.commitConfigUpdate('common-config', 'layouts/no-timer.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()


class TestTimerAlwaysDynamicBranches(ZuulTestCase):
    tenant_config_file = 'config/dynamic-only-project/dynamic.yaml'

    def test_exclude_always_dynamic_branches(self):
        # Test that no timer events are emitted for always dynamic branches.
        self.create_branch('org/project', 'stable')
        self.create_branch('org/project', 'feature/foo')
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'org/project', 'stable'))
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'org/project', 'feature/foo'))
        self.executor_server.hold_jobs_in_build = True
        self.commitConfigUpdate('common-config', 'layouts/timer.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))

        # The pipeline triggers every second, so we should have seen
        # several by now.
        for _ in iterate_timeout(60, 'jobs started'):
            if len(self.builds) > 1:
                break

        timer = self.scheds.first.sched.connections.drivers['timer']
        timer_jobs = timer.apsched.get_jobs()
        self.assertEqual(len(timer_jobs), 2)

        # Ensure that the status json has the ref so we can render it in the
        # web ui.
        manager = self.scheds.first.sched.abide.tenants[
            'tenant-one'].layout.pipeline_managers['periodic']
        self.assertEqual(len(manager.state.queues), 2)
        for queue in manager.state.queues:
            item = queue.queue[0]
            self.assertIn(item.changes[0].branch, ['master', 'stable'])

        self.executor_server.hold_jobs_in_build = False

        # Stop queuing timer triggered jobs so that the assertions
        # below don't race against more jobs being queued.
        self.commitConfigUpdate('common-config', 'layouts/no-timer.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))

        # If APScheduler is in mid-event when we remove the job, we
        # can end up with one more event firing, so give it an extra
        # second to settle.
        time.sleep(3)
        self.waitUntilSettled()

        self.executor_server.release()
        self.waitUntilSettled()

        self.assertTrue(len(self.history) > 1)
        for job in self.history[:1]:
            self.assertEqual(job.result, 'SUCCESS')
            self.assertEqual(job.name, 'project-bitrot')
            self.assertIn(job.ref, ('refs/heads/stable', 'refs/heads/master'))



# Copyright 2015 Rackspace Australia
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import re
from collections import OrderedDict
from urllib.parse import urlparse

import zuul.driver.zuul
import zuul.driver.gerrit
import zuul.driver.git
import zuul.driver.github
import zuul.driver.smtp
import zuul.driver.timer
import zuul.driver.sql
import zuul.driver.bubblewrap
import zuul.driver.nullwrap
import zuul.driver.mqtt
import zuul.driver.pagure
import zuul.driver.gitlab
import zuul.driver.elasticsearch
import zuul.driver.aws
import zuul.driver.azure
import zuul.driver.openstack
import zuul.driver.static
import zuul.driver.kubernetes
from zuul.connection import BaseConnection
from zuul.driver import (
    ProviderInterface,
    ReporterInterface,
    SourceInterface,
    TriggerInterface,
)
# Force subclass registration as a side-effect of importing it
from zuul.launcher.subnodes import SubnodeProviderNode  # noqa


class DefaultConnection(BaseConnection):
    pass


class ConnectionRegistry(object):
    """A registry of connections"""

    log = logging.getLogger("zuul.ConnectionRegistry")

    def __init__(self, check_bwrap=False):
        self.connections = OrderedDict()
        self.drivers = {}

        self.registerDriver(zuul.driver.zuul.ZuulDriver())
        self.registerDriver(zuul.driver.gerrit.GerritDriver())
        self.registerDriver(zuul.driver.git.GitDriver())
        self.registerDriver(zuul.driver.github.GithubDriver())
        self.registerDriver(zuul.driver.smtp.SMTPDriver())
        self.registerDriver(zuul.driver.timer.TimerDriver())
        self.registerDriver(zuul.driver.sql.SQLDriver())
        self.registerDriver(
            zuul.driver.bubblewrap.BubblewrapDriver(check_bwrap))
        self.registerDriver(zuul.driver.nullwrap.NullwrapDriver())
        self.registerDriver(zuul.driver.mqtt.MQTTDriver())
        self.registerDriver(zuul.driver.pagure.PagureDriver())
        self.registerDriver(zuul.driver.gitlab.GitlabDriver())
        self.registerDriver(zuul.driver.elasticsearch.ElasticsearchDriver())
        self.registerDriver(zuul.driver.aws.AwsDriver())
        self.registerDriver(zuul.driver.azure.AzureDriver())
        self.registerDriver(zuul.driver.openstack.OpenstackDriver())
        self.registerDriver(zuul.driver.static.StaticDriver())
        self.registerDriver(zuul.driver.kubernetes.KubernetesDriver())

    def registerDriver(self, driver):
        if driver.name in self.drivers:
            raise Exception("Driver %s already registered" % driver.name)
        self.drivers[driver.name] = driver

    def registerScheduler(self, sched):
        for driver_name, driver in self.drivers.items():
            driver.registerScheduler(sched)
        for connection_name, connection in self.connections.items():
            connection.registerScheduler(sched)

    def load(self, zk_client, component_registry):
        for connection in self.connections.values():
            connection.onLoad(zk_client, component_registry)

    def reconfigureDrivers(self, tenant):
        for driver in self.drivers.values():
            if hasattr(driver, 'reconfigure'):
                driver.reconfigure(tenant)

    def stop(self):
        for connection_name, connection in self.connections.items():
            connection.onStop()
        for driver in self.drivers.values():
            driver.stop()

    def configure(self, config, database=False, sources=False,
                  triggers=False, reporters=False, providers=False):
        # Register connections from the config
        connections = OrderedDict()

        if database and 'database' in config.sections():
            driver = self.drivers['sql']
            con_config = dict(config.items('database'))

            connection = driver.getConnection('database', con_config)
            connections['database'] = connection

        for section_name in config.sections():
            con_match = re.match(r'^connection ([\'\"]?)(.*)(\1)$',
                                 section_name, re.I)
            if not con_match:
                continue
            con_name = con_match.group(2)
            con_config = dict(config.items(section_name))

            if 'driver' not in con_config:
                raise Exception("No driver specified for connection %s."
                                % con_name)

            con_driver = con_config['driver']
            if (con_driver not in self.drivers) or con_driver == 'sql':
                raise Exception("Unknown driver, %s, for connection %s"
                                % (con_config['driver'], con_name))

            driver = self.drivers[con_driver]

            if (database and driver.name == 'sql' or
                sources and isinstance(driver, SourceInterface) or
                triggers and isinstance(driver, TriggerInterface) or
                reporters and isinstance(driver, ReporterInterface) or
                providers and isinstance(driver, ProviderInterface)):
                # Only create connections for the requested drivers
                connection = driver.getConnection(con_name, con_config)
                connections[con_name] = connection

        # If the [gerrit] or [smtp] sections still exist, load them in as a
        # connection named 'gerrit' or 'smtp' respectfully

        if 'gerrit' in config.sections():
            if 'gerrit' in connections:
                self.log.warning(
                    "The legacy [gerrit] section will be ignored in favour"
                    " of the [connection gerrit].")
            else:
                driver = self.drivers['gerrit']
                connections['gerrit'] = \
                    driver.getConnection(
                        'gerrit', dict(config.items('gerrit')))

        if 'smtp' in config.sections():
            if 'smtp' in connections:
                self.log.warning(
                    "The legacy [smtp] section will be ignored in favour"
                    " of the [connection smtp].")
            else:
                driver = self.drivers['smtp']
                connections['smtp'] = \
                    driver.getConnection(
                        'smtp', dict(config.items('smtp')))

        # Create default connections for drivers which need no
        # connection information (e.g., 'timer' or 'zuul').
        for driver in self.drivers.values():
            if not hasattr(driver, 'getConnection'):
                if (database and driver.name == 'sql' or
                    sources and isinstance(driver, SourceInterface) or
                    triggers and isinstance(driver, TriggerInterface) or
                    reporters and isinstance(driver, ReporterInterface) or
                    providers and isinstance(driver, ProviderInterface)):
                    connections[driver.name] = DefaultConnection(
                        driver, driver.name, {})

        if database:
            if 'database' not in connections:
                raise Exception("Database configuration is required")

        self.connections = connections

    def getSqlConnection(self):
        """
        Gets the SQL connection. This is either the connection
        described in the [database] section, or the first configured
        connection.

        :return: The SQL connection.

        """
        connection = self.connections.get('database')
        if not connection:
            raise Exception("No SQL connections")
        return connection

    def getSqlReporter(self, pipeline):
        """
        Gets the SQL reporter. Such reporter is based on
        `getSqlConnection`.

        :param pipeline: Pipeline
        :return: The SQL reporter

        """
        connection = self.getSqlConnection()
        return connection.driver.getReporter(connection, pipeline)

    def getSource(self, connection_name):
        connection = self.connections[connection_name]
        return connection.driver.getSource(connection)

    def getSources(self):
        sources = []
        for connection in self.connections.values():
            if hasattr(connection.driver, 'getSource'):
                sources.append(connection.driver.getSource(connection))
        return sources

    def getProviderConnections(self):
        return [c for c in self.connections.values()
                if isinstance(c.driver, ProviderInterface)]

    def getReporter(self, connection_name, pipeline, config=None,
                    parse_context=None):
        connection = self.connections[connection_name]
        return connection.driver.getReporter(
            connection, pipeline, config, parse_context)

    def getTrigger(self, connection_name, config=None):
        connection = self.connections[connection_name]
        return connection.driver.getTrigger(connection, config)

    def getTriggerEventClass(self, driver_name: str):
        driver = self.drivers[driver_name]
        return driver.getTriggerEventClass()

    def getSourceByHostname(self, hostname):
        for connection in self.connections.values():
            if hasattr(connection, 'canonical_hostname'):
                if connection.canonical_hostname == hostname:
                    return self.getSource(connection.connection_name)
            if hasattr(connection, 'server'):
                if connection.server == hostname:
                    return self.getSource(connection.connection_name)
            if hasattr(connection, 'baseurl'):
                if urlparse(connection.baseurl).hostname == hostname:
                    return self.getSource(connection.connection_name)
        return None

    def getSourceByCanonicalHostname(self, canonical_hostname):
        for connection in self.connections.values():
            if hasattr(connection, 'canonical_hostname'):
                if connection.canonical_hostname == canonical_hostname:
                    return self.getSource(connection.connection_name)
        return None



# Copyright 2022-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools

# This is a map of instance types to quota codes.  There does not
# appear to be an automated way to determine what quota code to use
# for an instance type, therefore this list was manually created by
# visiting
# https://us-west-1.console.aws.amazon.com/servicequotas/home/services/ec2/quotas
# and filtering by "Instances".  An example description is "Running
# On-Demand P instances" which we can infer means we should use that
# quota code for instance types starting with the letter "p".  All
# instance type names follow the format "([a-z\-]+)\d", so we can
# match the first letters (up to the first number) of the instance
# type name with the letters in the quota name.  The prefix "u-" for
# "Running On-Demand High Memory instances" was determined from
# https://aws.amazon.com/ec2/instance-types/high-memory/


ON_DEMAND = 0
SPOT = 1

INSTANCE_QUOTA_CODES = {
    # INSTANCE FAMILY: [ON-DEMAND, SPOT]
    'a': ['L-1216C47A', 'L-34B43A08'],
    'c': ['L-1216C47A', 'L-34B43A08'],
    'd': ['L-1216C47A', 'L-34B43A08'],
    'h': ['L-1216C47A', 'L-34B43A08'],
    'i': ['L-1216C47A', 'L-34B43A08'],
    'm': ['L-1216C47A', 'L-34B43A08'],
    'r': ['L-1216C47A', 'L-34B43A08'],
    't': ['L-1216C47A', 'L-34B43A08'],
    'z': ['L-1216C47A', 'L-34B43A08'],
    'dl': ['L-6E869C2A', 'L-85EED4F7'],
    'f': ['L-74FC7D96', 'L-88CF9481'],
    'g': ['L-DB2E81BA', 'L-3819A6DF'],
    'vt': ['L-DB2E81BA', 'L-3819A6DF'],
    'u-': ['L-43DA4232', ''],          # 'high memory'
    'inf': ['L-1945791B', 'L-B5D1601B'],
    'p': ['L-417A185B', 'L-7212CCBC'],
    'x': ['L-7295265B', 'L-E3A00192'],
    'trn': ['L-2C3B7624', 'L-6B0D517C'],
    'hpc': ['L-F7808C92', '']
}

HOST_QUOTA_CODES = {
    'a1': 'L-949445B0',
    'c3': 'L-8D142A2E',
    'c4': 'L-E4BF28E0',
    'c5': 'L-81657574',
    'c5a': 'L-03F01FD8',
    'c5d': 'L-C93F66A2',
    'c5n': 'L-20F13EBD',
    'c6a': 'L-D75D2E84',
    'c6g': 'L-A749B537',
    'c6gd': 'L-545AED39',
    'c6gn': 'L-5E3A299D',
    'c6i': 'L-5FA3355A',
    'c6id': 'L-1BBC5241',
    'c6in': 'L-6C2C40CC',
    'c7a': 'L-698B67E5',
    'c7g': 'L-13B8FCE8',
    'c7gd': 'L-EF58B059',
    'c7gn': 'L-97677CE3',
    'c7i': 'L-587AA6E3',
    'd2': 'L-8B27377A',
    'dl1': 'L-AD667A3D',
    'f1': 'L-5C4CD236',
    'g3': 'L-DE82EABA',
    'g3s': 'L-9675FDCD',
    'g4ad': 'L-FD8E9B9A',
    'g4dn': 'L-CAE24619',
    'g5': 'L-A6E7FE5E',
    'g5g': 'L-4714FFEA',
    'g6': 'L-B88B9D6B',
    'gr6': 'L-E68C3AFF',
    'h1': 'L-84391ECC',
    'i2': 'L-6222C1B6',
    'i3': 'L-8E60B0B1',
    'i3en': 'L-77EE2B11',
    'i4g': 'L-F62CBADB',
    'i4i': 'L-0300530D',
    'im4gn': 'L-93155D6F',
    'inf': 'L-5480EFD2',
    'inf2': 'L-E5BCF7B5',
    'is4gen': 'L-CB4F5825',
    'm3': 'L-3C82F907',
    'm4': 'L-EF30B25E',
    'm5': 'L-8B7BF662',
    'm5a': 'L-B10F70D6',
    'm5ad': 'L-74F41837',
    'm5d': 'L-8CCBD91B',
    'm5dn': 'L-DA07429F',
    'm5n': 'L-24D7D4AD',
    'm5zn': 'L-BD9BD803',
    'm6a': 'L-80F2B67F',
    'm6g': 'L-D50A37FA',
    'm6gd': 'L-84FB37AA',
    'm6i': 'L-D269BEFD',
    'm6id': 'L-FDB0A352',
    'm6idn': 'L-9721EDD9',
    'm6in': 'L-D037CF10',
    'm7a': 'L-4740F819',
    'm7g': 'L-9126620E',
    'm7gd': 'L-F8516154',
    'm7i': 'L-30E31217',
    'mac1': 'L-A8448DC5',
    'mac2': 'L-5D8DADF5',
    'mac2-m2': 'L-B90B5B66',
    'mac2-m2pro': 'L-14F120D1',
    'p2': 'L-2753CF59',
    'p3': 'L-A0A19F79',
    'p3dn': 'L-B601B3B6',
    'p4d': 'L-86A789C3',
    'p5': 'L-5136197D',
    'r3': 'L-B7208018',
    'r4': 'L-313524BA',
    'r5': 'L-EA4FD6CF',
    'r5a': 'L-8FE30D52',
    'r5ad': 'L-EC7178B6',
    'r5b': 'L-A2D59C67',
    'r5d': 'L-8814B54F',
    'r5dn': 'L-4AB14223',
    'r5n': 'L-52EF324A',
    'r6a': 'L-BC1589C5',
    'r6g': 'L-B6D6065D',
    'r6gd': 'L-EF284EFB',
    'r6i': 'L-F13A970A',
    'r6id': 'L-B89271A9',
    'r6idn': 'L-C4EABC2C',
    'r6in': 'L-EA99608B',
    'r7a': 'L-4D15192B',
    'r7g': 'L-67B8B4C7',
    'r7gd': 'L-01137DCE',
    'r7i': 'L-55E05032',
    'r7iz': 'L-BC9FCC71',
    't3': 'L-1586174D',
    'trn1': 'L-5E4FB836',
    'trn1n': 'L-39926A58',
    'u-12tb1': 'L-D6994875',
    'u-18tb1': 'L-5F7FD336',
    'u-24tb1': 'L-FACBE655',
    'u-3tb1': 'L-7F5506AB',
    'u-6tb1': 'L-89870E8E',
    'u-9tb1': 'L-98E1FFAC',
    'u7in-16tb': 'L-75B9BECB',
    'u7in-24tb': 'L-CA51381E',
    'u7in-32tb': 'L-9D28191F',
    'vt1': 'L-A68CFBF7',
    'x1': 'L-DE3D9563',
    'x1e': 'L-DEF8E115',
    'x2gd': 'L-5CC9EA82',
    'x2idn': 'L-A84ABF80',
    'x2iedn': 'L-D0AA08B1',
    'x2iezn': 'L-888B4496',
    'z1d': 'L-F035E935',
}

VOLUME_QUOTA_CODES = {
    'io1': dict(iops='L-B3A130E6', storage='L-FD252861'),
    'io2': dict(iops='L-8D977E7E', storage='L-09BD8365'),
    'sc1': dict(storage='L-17AF77E8'),
    'gp2': dict(storage='L-D18FCD1D'),
    'gp3': dict(storage='L-7A658B76'),
    'standard': dict(storage='L-9CF3C2EB'),
    'st1': dict(storage='L-82ACEF56'),
}

ALL_QUOTA_CODES = set(itertools.chain(
    (x for x in itertools.chain.from_iterable(
        INSTANCE_QUOTA_CODES.values()) if x),
    HOST_QUOTA_CODES.values(),
    itertools.chain.from_iterable(
        (x.values() for x in VOLUME_QUOTA_CODES.values())
    ),
))
